package dev.kske.chess.board;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dev.kske.chess.board.Piece.Color;
import dev.kske.chess.board.Piece.Type;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>Board.java</strong><br>
 * Created: <strong>01.07.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public class Board {

	private Piece[][]				boardArr	= new Piece[8][8];
	private Map<Color, Position>	kingPos		= new HashMap<>();
	private Log						log			= new Log();

	private static final Map<Type, int[][]> positionScores;

	static {
		positionScores = new HashMap<>();
		positionScores.put(Type.KING,
				new int[][] { new int[] { -3, -4, -4, -5, -5, -4, -4, -3 }, new int[] { -3, -4, -4, -5, -4, -4, -4, -3 },
						new int[] { -3, -4, -4, -5, -4, -4, -4, -3 }, new int[] { -3, -4, -4, -5, -4, -4, -4, -3 },
						new int[] { -2, -3, -3, -2, -2, -2, -2, -1 }, new int[] { -1, -2, -2, -2, -2, -2, -2, -1 },
						new int[] { 2, 2, 0, 0, 0, 0, 2, 2 }, new int[] { 2, 3, 1, 0, 0, 1, 3, 2 } });
		positionScores.put(Type.QUEEN,
				new int[][] { new int[] { -2, -1, -1, -1, -1, -1, -1, -2 }, new int[] { -1, 0, 0, 0, 0, 0, 0, -1 },
						new int[] { -1, 0, 1, 1, 1, 1, 0, -1 }, new int[] { -1, 0, 1, 1, 1, 1, 0, -1 }, new int[] { 0, 0, 1, 1, 1, 1, 0, -1 },
						new int[] { -1, 1, 1, 1, 1, 1, 0, -1 }, new int[] { -1, 0, 1, 0, 0, 0, 0, -1 },
						new int[] { -2, -1, -1, -1, -1, -1, -1, -2 } });
		positionScores.put(Type.ROOK,
				new int[][] { new int[] { 0, 0, 0, 0, 0, 0, 0, 0 }, new int[] { 1, 1, 1, 1, 1, 1, 1, 1 }, new int[] { -1, 0, 0, 0, 0, 0, 0, -1 },
						new int[] { -1, 0, 0, 0, 0, 0, 0, -1 }, new int[] { -1, 0, 0, 0, 0, 0, 0, -1 }, new int[] { -1, 0, 0, 0, 0, 0, 0, -1 },
						new int[] { -1, 0, 0, 0, 0, 0, 0, -1 }, new int[] { 0, 0, 0, 1, 1, 0, 0, 0 } });
		positionScores.put(Type.KNIGHT,
				new int[][] { new int[] { -5, -4, -3, -3, -3, -3, -4, -5 }, new int[] { -4, -2, 0, 0, 0, 0, -2, -4 },
						new int[] { -3, 0, 1, 2, 2, 1, 0, -3 }, new int[] { -3, 1, 2, 2, 2, 2, 1, -3 }, new int[] { -3, 0, 2, 2, 2, 2, 0, -1 },
						new int[] { -3, 1, 1, 2, 2, 1, 1, -3 }, new int[] { -4, -2, 0, 1, 1, 0, -2, -4 },
						new int[] { -5, -4, -3, -3, -3, -3, -4, -5 } });
		positionScores.put(Type.BISHOP,
				new int[][] { new int[] { -2, -1, -1, -1, -1, -1, -1, 2 }, new int[] { -1, 0, 0, 0, 0, 0, 0, -1 },
						new int[] { -1, 0, 1, 1, 1, 1, 0, -1 }, new int[] { -1, 1, 1, 1, 1, 1, 1, -1 }, new int[] { -1, 0, 1, 1, 1, 1, 0, -1 },
						new int[] { -1, 1, 1, 1, 1, 1, 1, -1 }, new int[] { -1, 1, 0, 0, 0, 0, 1, -1 },
						new int[] { -2, -1, -1, -1, -1, -1, -1, -2 } });
		positionScores.put(Type.PAWN,
				new int[][] { new int[] { 0, 0, 0, 0, 0, 0, 0, 0 }, new int[] { 5, 5, 5, 5, 5, 5, 5, 5 }, new int[] { 1, 1, 2, 3, 3, 2, 1, 1 },
						new int[] { 0, 0, 1, 3, 3, 1, 0, 0 }, new int[] { 0, 0, 0, 2, 2, 0, 0, 0 }, new int[] { 0, 0, -1, 0, 0, -1, 0, 0 },
						new int[] { 0, 1, 1, -2, -2, 1, 1, 0 }, new int[] { 0, 0, 0, 0, 0, 0, 0, 0 } });
	}

	/**
	 * Initializes the board with the default chess starting position.
	 */
	public Board() { initDefaultPositions(); }

	/**
	 * Creates a copy of another {@link Board} instance.<br>
	 * The created object is a deep copy, but does not contain any move history
	 * apart from the current {@link MoveNode}.
	 * 
	 * @param other The {@link Board} instance to copy
	 */
	public Board(Board other) {
		boardArr = new Piece[8][8];
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++) {
				if (other.boardArr[i][j] == null) continue;
				boardArr[i][j]			= (Piece) other.boardArr[i][j].clone();
				boardArr[i][j].board	= this;
			}

		kingPos.putAll(other.kingPos);
		log = new Log(other.log, false);
	}

	/**
	 * Moves a piece across the board if the move is legal.
	 * 
	 * @param move The move to execute
	 * @return {@code true}, if the attempted move was legal and thus executed
	 */
	public boolean attemptMove(Move move) {
		Piece piece = getPos(move);
		if (piece == null || !piece.isValidMove(move)) return false;
		else {
			// Set type after validation
			if (move.type == Move.Type.UNKNOWN) move.type = Move.Type.NORMAL;

			// Move piece
			move(move);

			// Revert move if it caused a check for its team
			if (checkCheck(piece.getColor())) {
				revert();
				return false;
			}

			return true;
		}
	}

	/**
	 * Moves a piece across the board without checking if the move is legal.
	 * 
	 * @param move The move to execute
	 */
	public void move(Move move) {
		Piece	piece			= getPos(move);
		Piece	capturePiece	= getDest(move);

		switch (move.type) {
			case PAWN_PROMOTION:
				setPos(move, null);
				// TODO: Select promotion
				setDest(move, new Queen(piece.getColor(), this));
				break;
			case EN_PASSANT:
				setDest(move, piece);
				setPos(move, null);
				boardArr[move.dest.x][move.dest.y - move.ySign] = null;
				break;
			case CASTLING:
				// Move the king
				setDest(move, piece);
				setPos(move, null);

				// Move the rook
				Move rookMove = move.dest.x == 6 ? new Move(7, move.pos.y, 5, move.pos.y) // Kingside
						: new Move(0, move.pos.y, 3, move.pos.y); // Queenside
				setDest(rookMove, getPos(rookMove));
				setPos(rookMove, null);
				break;
			case UNKNOWN:
				System.err.printf("Move of unknown type %s found!%n", move);
			case NORMAL:
				setDest(move, piece);
				setPos(move, null);
				break;
			default:
				System.err.printf("Move %s of unimplemented type found!%n", move);
		}

		// Update the king's position if the moved piece is the king
		if (piece.getType() == Type.KING) kingPos.put(piece.getColor(), move.dest);

		// Update log
		log.add(move, piece, capturePiece);
	}

	/**
	 * Moves a piece across the board without checking if the move is legal.
	 * 
	 * @param sanMove The move to execute in SAN (Standard Algebraic Notation)
	 */
	public void move(String sanMove) {
		Map<String, Pattern> patterns = new HashMap<>();
		patterns.put("pieceMove",
				Pattern.compile(
						"^(?<pieceType>[NBRQK])(?:(?<fromFile>[a-h])|(?<fromRank>[1-8])|(?<fromSquare>[a-h][1-8]))?x?(?<toSquare>[a-h][1-8])(?:\\+{0,2}|\\#)$"));
		patterns.put("pawnCapture",
				Pattern.compile("^(?<fromFile>[a-h])(?<fromRank>[1-8])?x(?<toSquare>[a-h][1-8])(?<promotedTo>[NBRQK])?(?:\\+{0,2}|\\#)?$"));
		patterns.put("pawnPush", Pattern.compile("^(?<toSquare>[a-h][1-8])(?<promotedTo>[NBRQK])?(?:\\+{0,2}|\\#)$"));
		patterns.put("castling", Pattern.compile("^(?<queenside>O-O-O)|(?<kingside>O-O)(?:\\+{0,2}|\\#)?$"));

		patterns.forEach((patternName, pattern) -> {
			Matcher m = pattern.matcher(sanMove);
			if (m.find()) {
				Position	pos			= null, dest = null;
				Move.Type	moveType	= Move.Type.NORMAL;
				switch (patternName) {
					case "pieceMove":
						dest = Position.fromLAN(m.group("toSquare"));
						if (m.group("fromSquare") != null) pos = Position.fromLAN(m.group("fromSquare"));
						else {
							Type	type	= Type.fromFirstChar(m.group("pieceType").charAt(0));
							char	file;
							int		rank;
							if (m.group("fromFile") != null) {
								file	= m.group("fromFile").charAt(0);
								rank	= get(type, file);
								pos		= Position.fromLAN(String.format("%c%d", file, rank));
							} else if (m.group("fromRank") != null) {
								rank	= Integer.parseInt(m.group("fromRank").substring(0, 1));
								file	= get(type, rank);
								pos		= Position.fromLAN(String.format("%c%d", file, rank));
							} else pos = get(type, dest);
						}
						break;
					case "pawnCapture":
						dest = Position.fromLAN(m.group("toSquare"));
						char file = m.group("fromFile").charAt(0);
						int rank = m.group("fromRank") == null ? get(Type.PAWN, file) : Integer.parseInt(m.group("fromRank"));
						pos = Position.fromLAN(String.format("%c%d", file, rank));
						break;
					case "pawnPush":
						dest = Position.fromLAN(m.group("toSquare"));
						// TODO: Pawn promotion
						int step = log.getActiveColor() == Color.WHITE ? 1 : -1;

						// One step forward
						if (boardArr[dest.x][dest.y + step] != null) pos = new Position(dest.x, dest.y + step);
						// Double step forward
						else pos = new Position(dest.x, dest.y + 2 * step);
						break;
					case "castling":
						pos = new Position(4, log.getActiveColor() == Color.WHITE ? 7 : 0);
						dest = new Position(m.group("kingside") != null ? 6 : 2, pos.y);
						moveType = Move.Type.CASTLING;
						break;
				}
				move(new Move(pos, dest, moveType));
				return;
			}
		});
	}

	/**
	 * Reverts the last move.
	 */
	public void revert() {
		MoveNode	moveNode		= log.getLast();
		Move		move			= moveNode.move;
		Piece		capturedPiece	= moveNode.capturedPiece;

		switch (move.type) {
			case PAWN_PROMOTION:
				setPos(move, new Pawn(getDest(move).getColor(), this));
				setDest(move, capturedPiece);
				break;
			case EN_PASSANT:
				setPos(move, getDest(move));
				setDest(move, null);
				boardArr[move.dest.x][move.dest.y - move.ySign] = new Pawn(getPos(move).getColor().opposite(), this);
				break;
			case CASTLING:
				// Move the king
				setPos(move, getDest(move));
				setDest(move, null);

				// Move the rook
				Move rookMove = move.dest.x == 6 ? new Move(5, move.pos.y, 7, move.pos.y) // Kingside
						: new Move(3, move.pos.y, 0, move.pos.y); // Queenside
				setDest(rookMove, getPos(rookMove));
				setPos(rookMove, null);
				break;
			case UNKNOWN:
				System.err.printf("Move of unknown type %s found!%n", move);
			case NORMAL:
				setPos(move, getDest(move));
				setDest(move, capturedPiece);
				break;
			default:
				System.err.printf("Move %s of unimplemented type found!%n", move);
		}

		// Update the king's position if the moved piece is the king
		if (getPos(move).getType() == Type.KING) kingPos.put(getPos(move).getColor(), move.pos);

		// Update log
		log.removeLast();
	}

	/**
	 * Generated every legal move for one color
	 * 
	 * @param color The color to generate the moves for
	 * @return A list of all legal moves
	 */
	public List<Move> getMoves(Color color) {
		List<Move> moves = new ArrayList<>();
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++)
				if (boardArr[i][j] != null && boardArr[i][j].getColor() == color) moves.addAll(boardArr[i][j].getMoves(new Position(i, j)));
		return moves;
	}

	public List<Move> getMoves(Position pos) { return get(pos).getMoves(pos); }

	/**
	 * Checks, if the king is in check.
	 * 
	 * @param color The color of the king to check
	 * @return {@code true}, if the king is in check
	 */
	public boolean checkCheck(Color color) {
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++) {
				Position pos = new Position(i, j);
				if (get(pos) != null && get(pos).getColor() != color && get(pos).isValidMove(new Move(pos, kingPos.get(color)))) return true;
			}
		return false;
	}

	/**
	 * Checks, if the king is in checkmate.
	 * This requires the king to already be in check!
	 * 
	 * @param color The color of the king to check
	 * @return {@code true}, if the king is in checkmate
	 */
	public boolean checkCheckmate(Color color) {
		// Return false immediately if the king can move
		if (!getMoves(kingPos.get(color)).isEmpty()) return false;
		else {
			for (Move move : getMoves(color)) {
				move(move);
				boolean check = checkCheck(color);
				revert();
				if (!check) return false;
			}
			return true;
		}
	}

	public GameState getGameEventType(Color color) {
		return checkCheck(color) ? checkCheckmate(color) ? GameState.CHECKMATE : GameState.CHECK
				: getMoves(color).isEmpty() || log.getLast().halfmoveClock >= 50 ? GameState.STALEMATE : GameState.NORMAL;
	}

	/**
	 * Evaluated the board.
	 * 
	 * @param color The color to evaluate for
	 * @return An positive number representing how good the position is
	 */
	public int evaluate(Color color) {
		int score = 0;
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++)
				if (boardArr[i][j] != null && boardArr[i][j].getColor() == color) {
					switch (boardArr[i][j].getType()) {
						case QUEEN:
							score += 90;
							break;
						case ROOK:
							score += 50;
							break;
						case KNIGHT:
							score += 30;
							break;
						case BISHOP:
							score += 30;
							break;
						case PAWN:
							score += 10;
							break;
					}
					if (positionScores.containsKey(boardArr[i][j].getType()))
						score += positionScores.get(boardArr[i][j].getType())[i][color == Color.WHITE ? j : 7 - j];
				}
		return score;
	}

	/**
	 * Initialized the board array with the default chess pieces and positions.
	 */
	public void initDefaultPositions() {
		// Initialize pawns
		for (int i = 0; i < 8; i++) {
			boardArr[i][1]	= new Pawn(Color.BLACK, this);
			boardArr[i][6]	= new Pawn(Color.WHITE, this);
		}

		// Initialize kings
		boardArr[4][0]	= new King(Color.BLACK, this);
		boardArr[4][7]	= new King(Color.WHITE, this);

		// Initialize king position objects
		kingPos.put(Color.BLACK, new Position(4, 0));
		kingPos.put(Color.WHITE, new Position(4, 7));

		// Initialize queens
		boardArr[3][0]	= new Queen(Color.BLACK, this);
		boardArr[3][7]	= new Queen(Color.WHITE, this);

		// Initialize rooks
		boardArr[0][0]	= new Rook(Color.BLACK, this);
		boardArr[0][7]	= new Rook(Color.WHITE, this);
		boardArr[7][0]	= new Rook(Color.BLACK, this);
		boardArr[7][7]	= new Rook(Color.WHITE, this);

		// Initialize knights
		boardArr[1][0]	= new Knight(Color.BLACK, this);
		boardArr[1][7]	= new Knight(Color.WHITE, this);
		boardArr[6][0]	= new Knight(Color.BLACK, this);
		boardArr[6][7]	= new Knight(Color.WHITE, this);

		// Initialize bishops
		boardArr[2][0]	= new Bishop(Color.BLACK, this);
		boardArr[2][7]	= new Bishop(Color.WHITE, this);
		boardArr[5][0]	= new Bishop(Color.BLACK, this);
		boardArr[5][7]	= new Bishop(Color.WHITE, this);

		// Clear all other tiles
		for (int i = 0; i < 8; i++)
			for (int j = 2; j < 6; j++)
				boardArr[i][j] = null;

		log.reset();
	}

	@Override
	public int hashCode() {
		final int	prime	= 31;
		int			result	= 1;
		result	= prime * result + Arrays.deepHashCode(boardArr);
		result	= prime * result + Objects.hash(kingPos, log);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (obj == null) return false;
		if (getClass() != obj.getClass()) return false;
		Board other = (Board) obj;
		return Arrays.deepEquals(boardArr, other.boardArr) && Objects.equals(kingPos, other.kingPos) && Objects.equals(log, other.log);
	}

	/**
	 * @param pos The position from which to return a piece
	 * @return The piece at the position
	 */
	public Piece get(Position pos) { return boardArr[pos.x][pos.y]; }

	/**
	 * Searches for a {@link Piece} inside a file (A - H).
	 * 
	 * @param type The {@link Type} of the piece to search for
	 * @param file The file in which to search for the piece
	 * @return The rank (1 - 8) of the first piece with the specified type and
	 *         current color in the file, or {@code -1} if there isn't any
	 */
	public int get(Type type, char file) {
		int x = file - 97;
		for (int i = 0; i < 8; i++)
			if (boardArr[x][i] != null && boardArr[x][i].getType() == type && boardArr[x][i].getColor() == log.getActiveColor()) return 8 - i;
		return -1;
	}

	/**
	 * Searches for a {@link Piece} inside a rank (1 - 8).
	 * 
	 * @param type The {@link Type} of the piece to search for
	 * @param rank The rank in which to search for the piece
	 * @return The file (A - H) of the first piece with the specified type and
	 *         current color in the file, or {@code -} if there isn't any
	 */
	public char get(Type type, int rank) {
		int y = rank - 1;
		for (int i = 0; i < 8; i++)
			if (boardArr[i][y] != null && boardArr[i][y].getType() == type && boardArr[i][y].getColor() == log.getActiveColor())
				return (char) (i + 97);
		return '-';
	}

	/**
	 * Searches for a {@link Piece} that can move to a {@link Position}.
	 * 
	 * @param type The {@link Type} of the piece to search for
	 * @param dest The destination that the piece is required to reach
	 * @return The position of a piece that can move to the specified destination
	 */
	public Position get(Type type, Position dest) {
		for (int i = 0; i < 8; i++)
			for (int j = 0; j < 8; j++)
				if (boardArr[i][j] != null && boardArr[i][j].getType() == type && boardArr[i][j].getColor() == log.getActiveColor()) {
					Position pos = new Position(i, j);
					if (boardArr[i][j].isValidMove(new Move(pos, dest))) return pos;
				}
		return null;
	}

	/**
	 * Places a piece at a position.
	 * 
	 * @param pos   The position to place the piece at
	 * @param piece The piece to place
	 */
	public void set(Position pos, Piece piece) { boardArr[pos.x][pos.y] = piece; }

	/**
	 * @param move The move from which position to return a piece
	 * @return The piece at the position of the move
	 */
	public Piece getPos(Move move) { return get(move.pos); }

	/**
	 * @param move The move from which destination to return a piece
	 * @return The piece at the destination of the move
	 */
	public Piece getDest(Move move) { return get(move.dest); }

	/**
	 * Places a piece at the position of a move.
	 * 
	 * @param move  The move at which position to place the piece
	 * @param piece The piece to place
	 */
	public void setPos(Move move, Piece piece) { set(move.pos, piece); }

	/**
	 * Places a piece at the destination of a move.
	 * 
	 * @param move  The move at which destination to place the piece
	 * @param piece The piece to place
	 */
	public void setDest(Move move, Piece piece) { set(move.dest, piece); }

	/**
	 * @return The board array
	 */
	public Piece[][] getBoardArr() { return boardArr; }

	/**
	 * @return The move log
	 */
	public Log getLog() { return log; }
}
